/**
 * Copyright 2012 Bryan Miller
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package lib.obsidianblk.midnightgdx;

import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.ShortBuffer;

import lib.obsidianblk.midnightgdx.exceptions.MeshException;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.utils.BufferUtils;


public class MeshVBO extends Mesh{
	public final static int BYTES_PER_FLOAT = 4;
	public final static int BYTES_PER_SHORT = 2;
	final static IntBuffer tmpHandle = BufferUtils.newIntBuffer(1);
	
	public final int ElementsPerVertex;
	public final int BytesPerVertex;
	
	private boolean strictExceptions = false;
	
	private Attributes meshAttributes = new Attributes();
	private int texDepth=2;
	
	private int vertexHandle = 0;
	private int indexHandle = 0;
	
	private final int vertexCount;
	private final FloatBuffer buffer;
	private final ShortBuffer ibuffer;
	
	private boolean dirtyv = true;
	private boolean dirtyi = true;
	private boolean bound = false;
	private boolean locked = false; // Mesh can no longer be edited.
	
	
	
	MeshVBO(int maxVerticies, int maxIndicies, int attribs, int textureDepth) throws MeshException {
		if (textureDepth < 1 || textureDepth > 3){
			throw new MeshException("Texture depth is out of bounds. Only 1D, 2D, or 3D textures allowed.");
		}
		if (!Attributes.ValidAttribValue(attribs, true)){
			throw new MeshException("Mesh attributes MUST include the Vertex attribute.");
		}
		meshAttributes.Support(attribs);
		
		// Normal, color, and texcoord counts are 0 if not included in buffer.
		int ncount = meshAttributes.Supports(Attributes.NORMAL) ? 3 : 0; // Normals are always defined in X, Y, Z... so a 3 count.
		int ccount = meshAttributes.Supports(Attributes.COLOR) ? 4 : 0; // Color is a count of 4 (R,G,B,A).
		// TexCoords have a number of elements depending on the dimension of the texture. Should usually be 2, but 1 and 3 are allowed... for now.
		int tcount = meshAttributes.Supports(Attributes.TEXCOORD) ? textureDepth : 0;
		texDepth = textureDepth;
		
		ElementsPerVertex = (3+ccount+ncount+tcount);
		BytesPerVertex = ElementsPerVertex*MeshVBO.BYTES_PER_FLOAT;
		buffer = BufferUtils.newUnsafeByteBuffer(BytesPerVertex * maxVerticies).asFloatBuffer();
		vertexHandle = CreateBufferHandle();
		
		ibuffer = BufferUtils.newUnsafeByteBuffer(maxIndicies*MeshVBO.BYTES_PER_SHORT).asShortBuffer();
		indexHandle = CreateBufferHandle();
		
		vertexCount = buffer.capacity()/ElementsPerVertex;
	}
	
	private int CreateBufferHandle(){
		Gdx.gl20.glGenBuffers(1, tmpHandle);
		return tmpHandle.get(0);
	}
	
	public boolean strictExceptions(){
		return strictExceptions;
	}
	
	public boolean strictExceptions(boolean strict){
		strictExceptions = strict;
		return strictExceptions;
	}
	
	public int vertexCount(){
		return vertexCount;
	}
	
	public int indexCount(){
		return ibuffer.capacity();
	}
	
	public boolean containsAttributes(int a){
		return meshAttributes.Supports(a);
	}
	
	public int getAttributes(){
		return meshAttributes.asInt();
	}
	
	public int texureDepth(){return texDepth;}
	
	public void finalize(){
		locked = true;
	}
	
	public boolean finalized(){
		return locked;
	}
	
	
	public void setIndicies(short[] src, int offset) throws MeshException{
		if (locked == false){
			int cpos = ibuffer.position();
			if (offset < 0 || offset >= ibuffer.capacity()){
				throw new MeshException("Buffer offset index is out of bounds.");
			}
			if (offset+src.length >= ibuffer.capacity()){
				throw new MeshException("Set request would exceed buffer capacity.");
			}
			ibuffer.position(offset);
			ibuffer.put(src, 0, src.length);
			ibuffer.position(cpos);
			dirtyi = true;
		} else if (strictExceptions){
			throw new MeshException("Cannot set data on a locked Mesh object.");
		}
	}
	
	public void setIndex(int index, short val) throws MeshException{
		if (locked == false){
			if (index < 0 || index >= ibuffer.capacity()){
				throw new MeshException("Index is out of bounds.");
			}
			ibuffer.put(index, val);
			dirtyi = true;
		} else if (strictExceptions){
			throw new MeshException("Cannot set data on a locked Mesh object.");
		}
	}
	
	public void setVerticies(float[] src, int offset) throws MeshException{
		if (locked == false){
			// Get the buffer's current position
			int cpos = buffer.position();
			// Confirm the src array is modulus the size of the vertex group.
			if (src.length%ElementsPerVertex != 0){
				throw new MeshException("Source array is not aligned to the size of a vertex group.");
			}
			
			// Calculate true buffer offset
			offset *= ElementsPerVertex;
			// Confirm the buffer is not out of bounds
			if (offset < 0 || offset >= buffer.capacity()){
				throw new MeshException("Buffer offset index is out of bounds.");
			}
			
			// Confirm the resulting array copy is not out of bounds.
			if (offset+src.length >= buffer.capacity()){
				throw new MeshException("Set request would exceed buffer capacity.");
			}
			
			// Offset
			buffer.position(offset);
			// Copy
			buffer.put(src, 0, src.length);
			// Reset to original buffer position... because we're nice like that.
			buffer.position(cpos);
			// Now that we've changed the buffer, we mark it as dirty, so that we reupload it to the graphics buffer.
			dirtyv = true;
		} else if (strictExceptions){
			throw new MeshException("Cannot set data on a locked Mesh object.");
		}
	}
	
	public void setVertexPosition(float[] pos, int offset) throws MeshException{
		if (locked == false){
			if (pos.length%3 != 0){
				throw new MeshException("Position array size mismatch. Expecting 3 components per vertex coord.");
			}
			if (offset < 0 || offset >= vertexCount){
				throw new MeshException("Vertex offset is out of bounds.");
			}
			if ((pos.length/3)+offset > vertexCount){
				throw new MeshException("Given vertex coord array is larger than vertex buffer given the current offset.");
			}
			int numcoords = pos.length/3;
			int cpos = buffer.position();
			offset *= ElementsPerVertex;
			
			for (int i = 0; i < numcoords; i++){
				buffer.position(offset + (i*ElementsPerVertex));
				buffer.put(pos, i*3, 3);
			}
			
			buffer.position(cpos);
			dirtyv = true;
		} else if (strictExceptions){
			throw new MeshException("Cannot set data on a locked Mesh object.");
		}
	}
	
	public void setVertexColor(float[] color, int offset) throws MeshException{
		int colordepth = 3;
		if (locked == false){
			if (meshAttributes.Supports(Attributes.COLOR)){
				if (color.length%colordepth != 0){
					colordepth = 4;
					if (color.length%colordepth != 0){
						throw new MeshException("Color array does not appear to support 24bit or 32bit color depths.");
					}
				}
				if (offset < 0 || offset >= vertexCount){
					throw new MeshException("Vertex index is out of bounds.");
				}
				if ((color.length/colordepth)+offset >= vertexCount){
					throw new MeshException("Given color array is larger than vertex buffer given the current offset.");
				}
				int elementcount = color.length/colordepth;
				int cpos = buffer.position();
				offset = (offset*ElementsPerVertex) + getElementOffset(1);
				
				for (int i = 0; i < elementcount; i++){
					buffer.position(offset + (i*ElementsPerVertex));
					buffer.put(color, i*colordepth, colordepth);
				}
				buffer.position(cpos);
			}
		} else if (strictExceptions){
			throw new MeshException("Cannot set data on a locked Mesh object.");
		}
	}
	
	public void setVertexNormal(float[] norm, int offset) throws MeshException{
		if (locked == false){
			if (meshAttributes.Supports(Attributes.NORMAL)){
				if (norm.length%3 != 0){
					throw new MeshException("Normal array must contain 3 items per element.");
				}
				if (offset < 0 || offset >= vertexCount){
					throw new MeshException("Vertex index is out of bounds.");
				}
				if ((norm.length/3)+offset >= vertexCount){
					throw new MeshException("Given normal array is larger than vertex buffer given the current offset.");
				}
				int elementcount = norm.length/3;
				int cpos = buffer.position();
				offset = (offset*ElementsPerVertex) + getElementOffset(2);
				
				for (int i = 0; i < elementcount; i++){
					buffer.position(offset + (i*elementcount));
					buffer.put(norm, i*3, 3);
				}
				buffer.position(cpos);
				dirtyv = true;
			} else if (strictExceptions){
				throw new MeshException("Mesh object does not contain Normal data.");
			}
		} else if (strictExceptions){
			throw new MeshException("Cannot set data on a locked Mesh object.");
		}
	}
	
	
	public void setVertexTexCoord(float[] texCoord, int offset) throws MeshException{
		if (locked == false){
			if (meshAttributes.Supports(Attributes.TEXCOORD)){
				if (texCoord.length%texDepth != 0){
					throw new MeshException("Texture Coordinate array size mismatch. Array does not match defined texture dimensions.");
				}
				if (offset < 0 || offset >= vertexCount){
					throw new MeshException("Vertex index is out of bounds.");
				}
				if ((texCoord.length/texDepth)+offset >= vertexCount){
					throw new MeshException("Given normal array is larger than vertex buffer given the current offset.");
				}
				int elementcount = texCoord.length/texDepth;
				int cpos = buffer.position();
				offset = (offset*ElementsPerVertex) + getElementOffset(3);
				
				for (int i = 0; i < elementcount; i++){
					buffer.position(offset+(i*ElementsPerVertex));
					buffer.put(texCoord, i*texDepth, texDepth);
				}
				buffer.position(cpos);
				dirtyv = true;
			}else if (strictExceptions){
				throw new MeshException("Mesh object does not contain Texture Coordinate data.");
			}
		} else if (strictExceptions){
			throw new MeshException("Cannot set data on a locked Mesh object.");
		}
	}
	
	public void bind(){
		if (!bound){
			buffer.position(0);
			Gdx.gl20.glBindBuffer(GL20.GL_ARRAY_BUFFER, vertexHandle);
			if (dirtyv){
				Gdx.gl20.glBufferData(GL20.GL_ARRAY_BUFFER, buffer.capacity()*MeshVBO.BYTES_PER_FLOAT, buffer, GL20.GL_STATIC_DRAW);
				dirtyv = false;
			}
			
			ibuffer.position(0);
			Gdx.gl20.glBindBuffer(GL20.GL_ELEMENT_ARRAY_BUFFER, indexHandle);
			if (dirtyi){
				Gdx.gl20.glBufferData(GL20.GL_ELEMENT_ARRAY_BUFFER, ibuffer.capacity()*MeshVBO.BYTES_PER_SHORT, ibuffer, GL20.GL_STATIC_DRAW);
				dirtyi = false;
			}
			bound = true;
		}
	}
	
	public void unbind(){
		if (bound){
			Gdx.gl20.glBindBuffer(GL20.GL_ARRAY_BUFFER, 0);
			Gdx.gl20.glBindBuffer(GL20.GL_ELEMENT_ARRAY_BUFFER, 0);
			bound = false;
		}
	}
	
	public void render(int glShape) throws MeshException{
		if (!bound){
			throw new MeshException("Cannot render. Mesh data must be bound first.");
		}
		Gdx.gl20.glDrawElements(glShape, ibuffer.capacity(), GL20.GL_UNSIGNED_SHORT, 0);
	}
	
	
	@Override
	public void dispose() {
		tmpHandle.clear();
		tmpHandle.put(vertexHandle);
		Gdx.gl20.glDeleteBuffers(1, tmpHandle);
		vertexHandle = 0;
		
		tmpHandle.clear();
		tmpHandle.put(indexHandle);
		Gdx.gl20.glDeleteBuffers(1, tmpHandle);
		indexHandle = 0;
		
		dirtyv = true;
		dirtyi = true;
	}
	

	private int getElementOffset(int EType){
		int offset = 0;
		switch(EType){
		case 3: // TexCoordInfo (Comes after normals)
			if (meshAttributes.Supports(Attributes.NORMAL)){
				offset += 3;
			}
		case 2: // Normals (Comes after colors)
			if (meshAttributes.Supports(Attributes.COLOR)){
				offset += 4;
			}
		case 1: // Colors (Comes after vertex position... which is ALWAYS part of the buffer)
			offset += 3;
		}
		return offset;
	}
}





