/**
 * Copyright 2012 Bryan Miller
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */



package lib.obsidianblk.midnightgdx;

import java.nio.Buffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;

import lib.obsidianblk.midnightgdx.exceptions.ShaderAttributeException;
import lib.obsidianblk.midnightgdx.exceptions.ShaderCompilerException;
import lib.obsidianblk.midnightgdx.exceptions.ShaderDefException;
import lib.obsidianblk.midnightgdx.exceptions.ShaderUniformException;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.GL20;


public class ShaderProgram extends Object{
	private Shader VertexShader = null;
	private Shader FragmentShader = null;
	
	private ArrayList<ShaderAttrib> Attribs = new ArrayList<ShaderAttrib>();
	private ArrayList<ShaderUniform> Uniforms = new ArrayList<ShaderUniform>();
	
	private int GLSLProgID = 0;
	
	private IntBuffer compileStatus = IntBuffer.allocate(1);
	
	
	/**
	 * Assigns the given Shader object as the vertex shader for the program.
	 * NOTE: Shader object must be defined as a GL_VERTEX_SHADER type, and the ShaderProgram must not already be compiled.
	 * 
	 * @param VShader - The Shader object being assigned (must be a GL_VERTEX_SHADER type).
	 * @throws ShaderDefException
	 */
	public void SetVertexShader(Shader VShader) throws ShaderDefException{
		if (GLSLProgID > 0){
			throw new ShaderDefException("Cannot store vertex shader: Shader program actively compiled and stored in the GLES state machine.");
		}
		if (VShader.GetShaderType() != GL20.GL_VERTEX_SHADER){
			throw new ShaderDefException("Cannot store vertex shader: Shader object is a fragment shader. Vertex shader required.");
		}
		VertexShader = VShader;
	}
	
	
	/**
	 * Assigns the given Shader object as the fragment shader for the program.
	 * NOTE: Shader object must be defined as a GL_FRAGMENT_SHADER type, and the ShaderProgram must not already be compiled.
	 * 
	 * @param FShader - The Shader object being assigned (must be a GL_FRAGMENT_SHADER type).
	 * @throws ShaderDefException
	 */
	public void SetFragmentShader(Shader FShader) throws ShaderDefException{
		if (GLSLProgID > 0){
			throw new ShaderDefException("Cannot store fragment shader: Shader program actively compiled and stored in the GLES state machine.");
		}
		if (FShader.GetShaderType() != GL20.GL_FRAGMENT_SHADER){
			throw new ShaderDefException("Cannot store fragment shader: Shader object is a vertex shader. Fragment shader required.");
		}
		FragmentShader = FShader;
	}
	
	/**
	 * Defines an attribute name and type that is assumed to be used by the shader program.
	 * WARNING: This method CANNOT be used AFTER a program has been compiled.
	 * @param Type - Must be one of Attributes.VERTEX, Attributes.NORMAL, Attributes.COLOR, or Attributes.TexCoord.
	 * @param AttributeName - Name of the attribute within the shader program. Attribute name MUST be unique.
	 * @throws ShaderAttributeException
	 */
	public void DefineAttribute(int Type, String AttributeName) throws ShaderAttributeException{
		if (GLSLProgID > 0){
			throw new ShaderAttributeException("Cannot define shader attributes after shader program has been compiled.");
		}
		if (!Attributes.OneOf(Type)){
			throw new ShaderAttributeException("Cannot use UNDEFINED attribute type when defining attributes.");
		}
		if (AttributeDefined(AttributeName)){
			throw new ShaderAttributeException("An attribute named \"" + AttributeName + "\" already defined.");
		}
		
		ShaderAttrib attrib = new ShaderAttrib(Type, Attribs.size(), AttributeName);
		Attribs.add(attrib);
	}
	
	
	/**
	 * Defines a uniform shader variable name that is assumed to be the MVP for the shader. Currently the ShaderProgram class requires that one of these
	 * types of uniform values exist in each program. MVPs are always assumed to be a FloatBuffer based uniform.
	 * NOTE: This method will throw an exception if an MVP uniform value was previously defined even if the name given is unique.
	 * @param size - The size of the matrix (2 for a 2x2. 3 for a 3x3. 4 for a 4x4)
	 * @param name - The uniform variable name within the Shader program. Must be unique.
	 * @throws ShaderUniformException
	 */
	public void DefineUniformMatrixName(int size, String name) throws ShaderUniformException{
		if (GLSLProgID > 0){
			throw new ShaderUniformException("Cannot define shader uniform matrix name after shader program has been compiled.");
		}
		if (size != 2 && size != 3 && size != 4){
			throw new ShaderUniformException("Uniform matrix size must be 2, 3, or 4.");
		}
		if (UniformDefined(name)){
			throw new ShaderUniformException("Uniform value '" + name + "' already defined.");
		}
		ShaderUniform s = new ShaderUniform(GL20.GL_FLOAT, name, size, true);
		Uniforms.add(s);
	}
	
	
	/**
	 * Defines a Uniform variable used in the Shader program.
	 * NOTE: type is defined using GL_* values, but they are interpreted as being their *Buffer object equivilent.
	 *   Ex. Defining a uniform as GL_FLOAT means that when passing the value in for the Uniform using the UniformPointer* methods, a FloatBuffer will
	 *   be expected.
	 * @param type - GL_FLOAT or GL_INT
	 * @param size - The number of elements in the uniform (from 1 to 4)
	 * @param name - The uniform variable name within the shader. Must be unique.
	 * @throws ShaderUniformException
	 */
	public void DefineUniform(int type, int size, String name) throws ShaderUniformException{
		if (GLSLProgID > 0){
			throw new ShaderUniformException("Cannot define shader uniform matrix name after shader program has been compiled.");
		}
		if (type != GL20.GL_FLOAT && type != GL20.GL_INT){
			throw new ShaderUniformException("Uniform buffer types can only be GL_FLOAT or GL_INT.");
		}
		if (size <= 0 || size > 4){
			throw new ShaderUniformException("Uniform variables can only contain a maximum of 4 elements.");
		}
		if (UniformDefined(name)){
			throw new ShaderUniformException("Uniform value '" + name + "' already defined.");
		}
		ShaderUniform s = new ShaderUniform(type, name, size);
		Uniforms.add(s);
	}
	
	/**
	 * Returns true if a uniform variable was defined with the given name and false otherwise.
	 * @param UniformName - Name of the uniform to find.
	 * @return boolean
	 */
	public boolean UniformDefined(String UniformName){
		for (ShaderUniform s: Uniforms){
			if (UniformName.compareTo(s.Name) == 0){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Returns true if the uniform variable with the given name is a Matrix. If the uniform variable is not a Matrix OR no uniform with the given
	 * UniformName is found, false is returned.
	 * @param UniformName - Name of the uniform variable to find.
	 * @return boolean
	 */
	public boolean UniformIsMatrix(String UniformName){
		for (ShaderUniform s: Uniforms){
			if (UniformName.compareTo(s.Name) == 0){
				return s.IsMatrix;
			}
		}
		return false;
	}
	
	private ShaderUniform GetUniformObj(String UniformName){
		for (ShaderUniform s: Uniforms){
			if (UniformName.compareTo(s.Name) == 0){
				return s;
			}
		}
		return null;
	}
	
	/**
	 * Returns true if an attribute was defined with the given AttributeName and false otherwise.
	 * @param AttributeName - Name of the attribute value to find.
	 * @return boolean
	 */
	public boolean AttributeDefined(String AttributeName){
		for (ShaderAttrib a : Attribs){
			if (AttributeName.compareTo(a.Name) == 0){
				return true;
			}
		}
		return false;
	}
	
	
	/**
	 * Returns true if an attribute was defined with the given AttributeType and false otherwise.
	 * @param AttributeType - The type of the attribute to find. Must be Attributes.VERTEX, Attributes.NORMAL, Attributes.COLOR, or Attribute.TEXCOORD
	 * @return boolean
	 */
	public boolean AttributeDefined(int AttributeType){
		if (Attributes.OneOf(AttributeType)){
			for (ShaderAttrib a : Attribs){
				if (a.Type == AttributeType){
					return true;
				}
			}
		}
		return false;
	}
	
	/**
	 * Returns the type value of the attribute with the name AttributeName. If no attribute is found with the given name, Attributes.UNDEFINED is
	 * returned.
	 * @param AttributeName - The name of the attribute in which to find.
	 * @return int
	 */
	public int AttributeType(String AttributeName){
		for (ShaderAttrib a : Attribs){
			if (AttributeName.compareTo(a.Name) == 0){
				return a.Type;
			}
		}
		return Attributes.UNDEFINED;
	}
	
	/**
	 * Returns the name of the attribute defined with the given Type. If no attribute of Type was defined, an empty String is returned.
	 * @param Type - The type of the attribute to find. Must be Attributes.VERTEX, Attributes.NORMAL, Attributes.COLOR, or Attributes.TEXCOORD.
	 * @return String
	 */
	public String AttributeName(int Type){
		for (ShaderAttrib a : Attribs){
			if (a.Type == Type){
				return a.Name;
			}
		}
		return "";
	}
	
	
	/**
	 * (WILL DESCRIBE LATER... Sorry)
	 * @param UniformName
	 * @param count
	 * @param transpose
	 * @param value
	 * @throws ShaderCompilerException
	 */
	public void UniformMatrixfv(String UniformName, int count, boolean transpose, FloatBuffer value) throws ShaderCompilerException{
		if (GLSLProgID <= 0){
			throw new ShaderCompilerException("Cannot pass uniform matrix buffer to uncompiled shader program.");
		}
		ShaderUniform s = GetUniformObj(UniformName);
		if (s == null){
			throw new ShaderCompilerException("Uniform Matrix is not defined.");
		}
		if (!s.IsMatrix){
			throw new ShaderCompilerException("Requested Uniform is not defined as a Matrix.");
		}
		int Handle = Gdx.gl20.glGetUniformLocation(GLSLProgID, s.Name);
		switch(s.Size){
		case 2:
			Gdx.gl20.glUniformMatrix2fv(Handle, count, transpose, value);
			break;
		case 3:
			Gdx.gl20.glUniformMatrix3fv(Handle, count, transpose, value);
			break;
		case 4:
			Gdx.gl20.glUniformMatrix4fv(Handle, count, transpose, value);
			break;
		}
	}
	
	/**
	 * (WILL DESCRIBE LATER... Sorry)
	 * @param UniformName
	 * @param count
	 * @param value
	 * @throws ShaderCompilerException
	 */
	public void Uniformiv(String UniformName, int count, IntBuffer value) throws ShaderCompilerException{
		if (GLSLProgID <= 0){
			throw new ShaderCompilerException("Cannot pass uniform buffer to uncompiled shader program.");
		}
		ShaderUniform s = GetUniformObj(UniformName);
		if (s == null){
			throw new ShaderCompilerException("Uniform variable is not defined.");
		}
		if (s.IsMatrix){
			throw new ShaderCompilerException("Requested Uniform is defined as a Matrix.");
		}
		if (s.Type != GL20.GL_INT){
			throw new ShaderCompilerException("Attempting to give a defined NON-INT uniform a INT buffer.");
		}
		int Handle = Gdx.gl20.glGetUniformLocation(GLSLProgID, s.Name);
		switch(s.Size){
		case 1:
			Gdx.gl20.glUniform1iv(Handle, count, value);
			break;
		case 2:
			Gdx.gl20.glUniform2iv(Handle, count, value);
			break;
		case 3:
			Gdx.gl20.glUniform3iv(Handle, count, value);
			break;
		case 4:
			Gdx.gl20.glUniform4iv(Handle, count, value);
			break;
		}
	}
	
	
	public void Uniformi(String UniformName, int v1) throws ShaderCompilerException{
		Uniformi(UniformName, v1, 0, 0, 0);
	}
	
	public void Uniformi(String UniformName, int v1, int v2) throws ShaderCompilerException{
		Uniformi(UniformName, v1, v2, 0, 0);
	}
	
	public void Uniformi(String UniformName, int v1, int v2, int v3) throws ShaderCompilerException{
		Uniformi(UniformName, v1, v2, v3, 0);
	}
	
	public void Uniformi(String UniformName, int v1, int v2, int v3, int v4) throws ShaderCompilerException{
		if (GLSLProgID <= 0){
			throw new ShaderCompilerException("Cannot pass uniform buffer to uncompiled shader program.");
		}
		ShaderUniform s = GetUniformObj(UniformName);
		if (s == null){
			throw new ShaderCompilerException("Uniform variable is not defined.");
		}
		if (s.IsMatrix){
			throw new ShaderCompilerException("Requested Uniform is defined as a Matrix.");
		}
		if (s.Type != GL20.GL_INT){
			throw new ShaderCompilerException("Attempting to give a defined NON-INT uniform an INT value.");
		}
		int Handle = Gdx.gl20.glGetUniformLocation(GLSLProgID, s.Name);
		switch(s.Size){
		case 1:
			Gdx.gl20.glUniform1i(Handle, v1);
			break;
		case 2:
			Gdx.gl20.glUniform2i(Handle, v1, v2);
			break;
		case 3:
			Gdx.gl20.glUniform3i(Handle, v1, v2, v3);
			break;
		case 4:
			Gdx.gl20.glUniform4i(Handle, v1, v2, v3, v4);
			break;
		}
	}
	
	/**
	 * (WILL DESCRIBE LATER... Sorry)
	 * @param UniformName
	 * @param count
	 * @param value
	 * @throws ShaderCompilerException
	 */
	public void Uniformfv(String UniformName, int count, FloatBuffer value) throws ShaderCompilerException{
		if (GLSLProgID <= 0){
			throw new ShaderCompilerException("Cannot pass uniform buffer to uncompiled shader program.");
		}
		ShaderUniform s = GetUniformObj(UniformName);
		if (s == null){
			throw new ShaderCompilerException("Uniform variable is not defined.");
		}
		if (s.IsMatrix){
			throw new ShaderCompilerException("Requested Uniform is defined as a Matrix.");
		}
		if (s.Type != GL20.GL_FLOAT){
			throw new ShaderCompilerException("Attempting to give a defined NON-FLOAT uniform a FLOAT buffer.");
		}
		int Handle = Gdx.gl20.glGetUniformLocation(GLSLProgID, s.Name);
		switch(s.Size){
		case 1:
			Gdx.gl20.glUniform1fv(Handle, count, value);
			break;
		case 2:
			Gdx.gl20.glUniform2fv(Handle, count, value);
			break;
		case 3:
			Gdx.gl20.glUniform3fv(Handle, count, value);
			break;
		case 4:
			Gdx.gl20.glUniform4fv(Handle, count, value);
			break;
		}
	}
	
	
	public void Uniformf(String UniformName, float v1) throws ShaderCompilerException{
		Uniformf(UniformName, v1, 0.0f, 0.0f, 0.0f);
	}
	
	public void Uniformf(String UniformName, float v1, float v2) throws ShaderCompilerException{
		Uniformf(UniformName, v1, v2, 0.0f, 0.0f);
	}
	
	public void Uniformf(String UniformName, float v1, float v2, float v3) throws ShaderCompilerException{
		Uniformf(UniformName, v1, v2, v3, 0.0f);
	}
	
	public void Uniformf(String UniformName, float v1, float v2, float v3, float v4) throws ShaderCompilerException{
		if (GLSLProgID <= 0){
			throw new ShaderCompilerException("Cannot pass uniform buffer to uncompiled shader program.");
		}
		ShaderUniform s = GetUniformObj(UniformName);
		if (s == null){
			throw new ShaderCompilerException("Uniform variable is not defined.");
		}
		if (s.IsMatrix){
			throw new ShaderCompilerException("Requested Uniform is defined as a Matrix.");
		}
		if (s.Type != GL20.GL_FLOAT){
			throw new ShaderCompilerException("Attempting to give a defined NON-FLOAT uniform a FLOAT value.");
		}
		int Handle = Gdx.gl20.glGetUniformLocation(GLSLProgID, s.Name);
		switch(s.Size){
		case 1:
			Gdx.gl20.glUniform1f(Handle, v1);
			break;
		case 2:
			Gdx.gl20.glUniform2f(Handle, v1, v2);
			break;
		case 3:
			Gdx.gl20.glUniform3f(Handle, v1, v2, v3);
			break;
		case 4:
			Gdx.gl20.glUniform4f(Handle, v1, v2, v3, v4);
			break;
		}
	}
	
	
	/**
	 * Assigns a shader attribute defined by the given AttributeName to the passed in Buffer object.
	 * NOTE: Attribute with AttributeName must be defined for a Buffer to be pointed to it.
	 * 
	 * @param AttributeName - Name of the attribute being assigned the Buffer pointer.
	 * @param size - Number of elements in attribute.
	 * @param type - The data type as defined by OpenGL (Ex: GL_FLOAT)
	 * @param normalized - Is the data normalized
	 * @param stride - How many bytes between each set of elements.
	 * @param ptr - The Buffer from which the attribute data is obtained.
	 * @throws ShaderCompilerException
	 */
	public void AttribPointer(String AttributeName, int size, int type, boolean normalized, int stride, Buffer ptr) throws ShaderCompilerException{
		if (GLSLProgID <= 0){
			throw new ShaderCompilerException("Cannot define vertex attribute pointer with an uncompiled shader program.");
		}
		if (!AttributeDefined(AttributeName)){
			throw new ShaderCompilerException("Attribute name \"" + AttributeName + "\" is not defined for this shader program.");
		}
		int handle = Gdx.gl20.glGetAttribLocation(GLSLProgID, AttributeName);
		Gdx.gl20.glVertexAttribPointer(handle, size, type, normalized, stride, ptr);
		Gdx.gl20.glEnableVertexAttribArray(handle);
	}
	
	/**
	 * Assigns the attribute of the given Type to the passed in Buffer object.
	 * NOTE: Attribute of the given Type must be defined for a Buffer to be pointed to it.
	 * 
	 * @param Type - The AttribType of the attribute (VERTEX, COLOR, NORMAL, TEXCOORD).
	 * @param size - Number of elements in attribute.
	 * @param type - The data type as defined by OpenGL (Ex: GL_FLOAT)
	 * @param normalized - Is the data normalized
	 * @param stride - How many bytes between each set of elements.
	 * @param ptr - The Buffer from which the attribute data is obtained.
	 * @throws ShaderCompilerException
	 */
	public void AttribPointer(int Type, int size, int type, boolean normalized, int stride, Buffer ptr) throws ShaderCompilerException{
		String AttributeName = "";
		if (GLSLProgID <= 0){
			throw new ShaderCompilerException("Cannot define vertex attribute pointer with an uncompiled shader program.");
		}
		AttributeName = AttributeName(Type);
		if (AttributeName.compareTo("") == 0){
			throw new ShaderCompilerException("Attempting to point to undefined attribute type.");
		}
		int handle = Gdx.gl20.glGetAttribLocation(GLSLProgID, AttributeName);
		Gdx.gl20.glVertexAttribPointer(handle, size, type, normalized, stride, ptr);
		Gdx.gl20.glEnableVertexAttribArray(handle);
	}
	
	public int Compile() throws ShaderCompilerException{
		if (GLSLProgID > 0){
			return GLSLProgID;
		} else if (!AttributeDefined(Attributes.VERTEX)){
			throw new ShaderCompilerException("No VERTEX shader attribute defined.");
		}
		
		if (!VertexShader.IsCompiled()){
			VertexShader.Compile();
		}
		if (!FragmentShader.IsCompiled()){
			FragmentShader.Compile();
		}
		
		// Create a program object and store the handle to it.
		GLSLProgID = Gdx.gl20.glCreateProgram();
		 
		if (GLSLProgID != 0)
		{
		    // Bind the vertex shader to the program.
			Gdx.gl20.glAttachShader(GLSLProgID, VertexShader.GetShaderGLID());
		 
		    // Bind the fragment shader to the program.
			Gdx.gl20.glAttachShader(GLSLProgID, FragmentShader.GetShaderGLID());
		 
		    // Bind attributes
			for (ShaderAttrib a : Attribs){
				Gdx.gl20.glBindAttribLocation(GLSLProgID, a.ID, a.Name);
			}
		 
		    // Link the two shaders together into a program.
			Gdx.gl20.glLinkProgram(GLSLProgID);
		 
		    // Get the link status.
		    Gdx.gl20.glGetProgramiv(GLSLProgID, GL20.GL_LINK_STATUS, compileStatus);
		 
		    // If the link failed, delete the program.
		    if (compileStatus.get(0) == 0)
		    {
		    	Gdx.gl20.glDeleteProgram(GLSLProgID);
		        GLSLProgID = 0;
		        throw new ShaderCompilerException("Error compiling shader program.");
		    }
		}
		
		return GLSLProgID;
	}
	
	public void Decompile(){
		if (GLSLProgID > 0){
			Gdx.gl20.glDeleteProgram(GLSLProgID);
			GLSLProgID = 0;
			
			VertexShader.Decompile();
			FragmentShader.Decompile();
		}
	}
	
	/**
	 * Returns true if the shader program is compiled and false otherwise.
	 * @return boolean
	 */
	public boolean IsCompiled(){return (GLSLProgID>0);}
	
	/**
	 * Sets this shader program to be the active shader program. This call will override any previous ShaderProgram::use() call.
	 * @throws ShaderCompilerException
	 */
	public void use() throws ShaderCompilerException{
		if (GLSLProgID > 0 && (!VertexShader.IsCompiled() || !FragmentShader.IsCompiled())){
			Decompile(); // Release the shader program ID... (Probably already invalidated)
			Compile(); // Recompile the shader... we want to USE it for Fates' sake!
		}
		Gdx.gl20.glUseProgram(GLSLProgID);
	}
	
	public static void UnuseShaders(){
		Gdx.gl20.glUseProgram(0);
	}
	
	
	private class ShaderAttrib{
		public int Type = 0;
		public int ID = -1;
		public String Name = "";
		
		//public ShaderAttrib(){}
		public ShaderAttrib(int t, int id, String n){
			Type = t;
			ID = id;
			Name = n;
		}
	}
	
	private class ShaderUniform{
		public boolean IsMatrix = false;
		public int Size = 0;
		public int Type = 0;
		public String Name = "";
		
		public ShaderUniform(int t, String n, int s){
			Type = t;
			Name = n;
			Size = s;
			IsMatrix = false;
		}
		
		public ShaderUniform(int t, String n, int s, boolean m){
			Size = s;
			Type = t;
			Name = n;
			IsMatrix = m;
		}
		//public ShaderUniform(){}
	}
}





